#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import os, sys, glob, re
import numpy as np
import matplotlib.pyplot as plt
import scipy.integrate
import scipy.special
import scipy.interpolate
from scipy.io import FortranFile
import scipy.fft as FFT

from mpi4py import MPI
comm = MPI.COMM_WORLD
rank = comm.Get_rank()

from .read_upf import UPF
AUTOA = 0.529177249  # VASP A.U. to Angstrom
# Unitility functions to build atomic wfc from the pseudopotential and build the projections U (and overlaps if needed)

def read_pswfc_from_upf(data_controller, atom):
  arry, attr = data_controller.data_dicts()
  
  # loop over species
  for (at, pseudo) in arry['species']:
    if atom == at:
      upf = UPF(os.path.join(attr['fpath'], pseudo))
      arry['jchia'][at] = upf.jchia
      return upf.r, upf.pswfc, pseudo
    
  raise RuntimeError('atom not found')
  
def radialfft_simpson(r, f, l, qmesh, volume):
  fq = np.zeros_like(qmesh)
  fact = 4.0*np.pi / np.sqrt(volume)
  
  f[r > 10.0] = 0.0
  for iq in range(len(qmesh)):
    q = qmesh[iq]
    bess = scipy.special.spherical_jn(l, r*q)
    aux = f * bess * r
    fq[iq] = scipy.integrate.simpson(aux, r)*fact
  return fq



def assign_jm(basis):
    ib = 0
    while ib < len(basis):
        b = basis[ib]
        l, m = b['l'], b['m']
        
        if l == 0:
            for c,jm in enumerate((0,1)):
                basis[ib+c]['jm'] = jm
            ib += 2

        elif l == 1:
            for c,jm in enumerate((2,3,4,5,6,7,)):
                basis[ib+c]['jm'] = jm
            ib += 6
            
        elif l == 2:
            for c,jm in enumerate((8,9,10,11,12,13,14,15,16,17)):
                basis[ib+c]['jm'] = jm
            ib += 10
            
        elif l == 3:
            for c,jm in enumerate((18,19,20,21,22,23,24,25,26,27,28,29,30,31)):
                basis[ib+c]['jm'] = jm
            ib += 14
        
        else:
            raise NotImplemented('l > 3 not implemented')


def build_pswfc_basis_all(data_controller):
  arry, attr = data_controller.data_dicts()
  verbose = attr['verbose']
  
  # build the mesh in q space
  ecutrho = attr['ecutrho']
  dq = 0.01
  qmesh = np.arange(0, np.sqrt(ecutrho) + 4, dq)
  volume = attr['omega']
  
  # loop over atoms
  basis,shells = [],{}
  arry['jchia'] = {}
  for na in range(len(arry['atoms'])):
    atom = arry['atoms'][na]
    tau = arry['tau'][na]
    r, pswfc, pseudo = read_pswfc_from_upf(data_controller, atom)
    if verbose and rank == 0:
      print('atom: {0:2s}  pseudo: {1:30s}  tau: {2}'.format(atom, pseudo, tau))
      
    # loop over pswfc'c
    a_shells = []
    jchia = []

    s = 0.5
    for pao in pswfc:
      l = 'SPDF'.find(pao['label'][1].upper())
      assert l != -1

      a_shells.append(l)
      if l == 0:
         jchia.append(0.5)
         s = 0.5
      else:
         jchia.append(l-s)
         s = -s

      wfc_g = radialfft_simpson(r, pao['wfc'], l, qmesh, volume)
      
      for m in range(1, 2*l+2):
        basis.append({'atom': atom, 'tau': tau, 'l': l, 'm': m, 'label': pao['label'],
          'r': r, 'wfc': pao['wfc'].copy(), 'qmesh': qmesh, 'wfc_g': wfc_g})
        if verbose and rank == 0:
          print('      atwfc: {0:3d}  {3}  l={1:d}, m={2:-d}'.format(len(basis), l, m, pao['label']))

      if attr['dftSO']:
        if l == 0:
            basis.append({'atom': atom, 'tau': tau, 'l': l, 'm': m, 'label': pao['label'],
              'r': r, 'wfc': pao['wfc'].copy(), 'qmesh': qmesh, 'wfc_g': wfc_g})
            if verbose and rank == 0:
              print('      atwfc: {0:3d}  {3}  l={1:d}, m={2:-d}'.format(len(basis), l, m, pao['label']))

    if atom not in shells:
      shells[atom] = a_shells
      arry['jchia'][atom] = jchia
    
  # in case of SO: assign the jm index
  if attr['dftSO']:
    assign_jm(basis)
    #for b in basis:
    #    l,m,jm = b['l'], b['m'], b['jm']
    #    print(l,m,jm)

  return basis,shells



def build_aewfc_basis(data_controller):
  arry, attr = data_controller.data_dicts()
  verbose = attr['verbose']
  
  # read the atomic bases
  aebasis = []
  for na in range(len(arry['atoms'])):
    atom_label = arry['atoms'][na]  # CHANGED
    elem = re.split('\d+', atom_label)[0]
    aefiles = glob.glob(attr['basispath']+str(elem)+'/*.dat')
    label = []
    for entry in aefiles:
      # label.append(entry.split('/')[-1].split('.')[0])
      label.append(entry[-6:-4])  
    aebasis.append(dict(zip(label,aefiles)))
  
  # build the mesh in q space
  ecutrho = attr['ecutrho']
  dq = 0.01
  qmesh = np.arange(0, np.sqrt(ecutrho) + 4, dq)
  volume = attr['omega']
  
  # loop over atoms
  basis,shells = [],{}
  arry['jchia'] = {}
  for na in range(len(arry['atoms'])):
    atom = arry['atoms'][na]
    elem = re.split('\d+', atom)[0]
    tau = arry['tau'][na]
    aewfc = []
    for shell in arry['configuration'][elem]:
      data = np.loadtxt(aebasis[na][shell])
      aewfc.append({shell : data[:,1], 'r' : data[:,0]})
      
      if verbose and rank == 0:
        print('atom: {0:2s}  AEWFC: {1:30s}  tau: {2}'.format(atom, aebasis[na][shell], tau))

    ash = []
    jchia = []
    for n in range(len(aewfc)):
      l = 'SPDF'.find(list(aewfc[n].items())[0][0][1].upper())
      assert l != -1
      ash.append(l)

      if attr['dftSO']:
         if l == 0:
             jchia.append(0.5)
         else:
             ash.append(l)
             jchia.append(l-0.5)
             jchia.append(l+0.5)

      wfc_g = radialfft_simpson(aewfc[n]['r'], aewfc[n][list(aewfc[n].items())[0][0]], l, qmesh, volume)

      twice = 1
      if attr['dftSO']: twice = 2

      for _ in range(twice):
        for m in range(1, 2*l+2):
          basis.append({'atom': atom, 'tau': tau, 'l': l, 'm': m, 'label': list(aewfc[n].items())[0][0],
                        'r': aewfc[n]['r'], 'wfc': aewfc[n][list(aewfc[n].items())[0][0]].copy(), 
                        'qmesh': qmesh, 'wfc_g': wfc_g})
          if verbose and rank == 0:
              print('      atwfc: {0:3d}  {3}  l={1:d}, m={2:-d}'.format(len(basis), l, m, list(aewfc[n].items())[0][0]))

    if atom not in shells:
      shells[atom] = ash
      arry['jchia'][atom] = jchia
    
  # in case of SO: assign the jm index
  if attr['dftSO']:
    assign_jm(basis)
    #for b in basis:
    #    l,m,jm = b['l'], b['m'], b['jm']
    #    print(l,m,jm)

  return basis,shells



def fft_wfc_G2R_old(wfc, igwx, gamma_only, mill, nr1, nr2, nr3, omega):
  wfcg = np.zeros((nr1,nr2,nr3), dtype=complex)
  
  for ig in range(igwx):
    wfcg[mill[0,ig],mill[1,ig],mill[2,ig]] = wfc[ig]
    if gamma_only:
      wfcg[-mill[0,ig],-mill[1,ig],-mill[2,ig]] = np.conj(wfc[ig])
      
  wfcr = FFT.ifftn(wfcg) * nr1 * nr2 * nr3 / np.sqrt(omega)
  return wfcr

def fft_wfc_G2R(wfc, gkspace, nr1, nr2, nr3, omega):
  igwx = gkspace['igwx']
  gamma_only = gkspace['gamma_only']
  mill = gkspace['mill']
  wfcg = np.zeros((nr1,nr2,nr3), dtype=complex)
  
  for ig in range(igwx):
    wfcg[mill[0,ig],mill[1,ig],mill[2,ig]] = wfc[ig]
    if gamma_only:
      wfcg[-mill[0,ig],-mill[1,ig],-mill[2,ig]] = np.conj(wfc[ig])
      
  wfcr = FFT.ifftn(wfcg) * nr1 * nr2 * nr3 / np.sqrt(omega)
  return wfcr

def fft_allwfc_G2R(wfc, gkspace, nr1, nr2, nr3, omega):
  igwx = gkspace['igwx']
  gamma_only = gkspace['gamma_only']
  mill = gkspace['mill']
  try: 
    nox = wfc.shape[0]
    nwx = wfc.shape[1]
    wfcg = np.zeros((nwx,nr1,nr2,nr3), dtype=complex)
  except:
    nox = 0
    wfcg = np.zeros((nr1,nr2,nr3), dtype=complex)
  if nox == 0:
    for ig in range(igwx):
      wfcg[:,mill[0,ig],mill[1,ig],mill[2,ig]] = wfc[:,ig]
      if gamma_only:
        wfcg[:,-mill[0,ig],-mill[1,ig],-mill[2,ig]] = np.conj(wfc[:,ig])
    wfcr = FFT.ifftn(wfcg) * nr1 * nr2 * nr3 / np.sqrt(omega)
  else:
    wfcr = np.zeros((nox,nr1,nr2,nr3), dtype=complex)
    for no in range(nox):
      for ig in range(igwx):
        wfcg[no,mill[0,ig],mill[1,ig],mill[2,ig]] = wfc[no,ig]
        if gamma_only:
          wfcg[no,-mill[0,ig],-mill[1,ig],-mill[2,ig]] = np.conj(wfc[no,ig])
      wfcr[no] = FFT.ifftn(wfcg[no]) * nr1 * nr2 * nr3 / np.sqrt(omega)
  return wfcr
  
def fft_wfc_R2G(wfc, igwx, mill, omega):
  tmp = FFT.fftn(wfc) / np.sqrt(omega)
  
  wfcg = np.zeros((igwx,), dtype=complex)
  for ig in range(igwx):
    wfcg[tmp] = tmp[mill[0,ig],mill[1,ig],mill[2,ig]]
    
  return wfcg

def read_QE_wfc(data_controller, ik, ispin):
  arry, attr = data_controller.data_dicts()
  
  if attr['nspin'] == 1 or attr['nspin'] == 4:
    wfcfile = 'wfc{0}.dat'.format(ik+1)
  elif attr['nspin'] == 2 and ispin == 0:
    wfcfile = 'wfcdw{0}.dat'.format(ik+1)
  elif attr['nspin'] == 2 and ispin == 1:
    wfcfile = 'wfcup{0}.dat'.format(ik+1)
  else:
    print('no wfc file found')
    
  with FortranFile(os.path.join(attr['fpath'], wfcfile), 'r') as f:
    record = f.read_ints(np.int32)
    assert len(record) == 11, 'something wrong reading fortran binary file'
    
    ik_ = record[0]
    assert ik+1 == ik_, 'wrong k-point in wfc file???'
    
    xk = np.frombuffer(record[1:7], np.float64)
    ispin = record[7]
    gamma_only = (record[8] != 0)
    scalef = np.frombuffer(record[9:], np.float64)[0]
    #print('ik =', ik, '  ispin =', ispin, '  gamma_only =', gamma_only, '  scalef =', scalef)
    #print('xk =', xk)
    
    ngw, igwx, npol, nbnd = f.read_ints(np.int32)
    #print('ngw =', ngw, '  igwx = ', igwx, '  npol = ', npol, '  nbnd = ', nbnd)
    bg = f.read_reals(np.float64).reshape(3,3,order='F')
    #print('bg = ', bg)
    mill = f.read_ints(np.int32).reshape(3,igwx,order='F')
    #print('mill.shape = ', mill.shape)
    
    wfc = []
    for i in range(nbnd):
      wfc.append(f.read_reals(np.complex128))
  
  # compute overlap
  ovp = np.zeros((nbnd,nbnd),dtype=complex)
  for n in range(nbnd):
    for m in range(nbnd):
      ovp[n,m] = np.sum(np.conj(wfc[n]).dot(wfc[m]))
  eigs, eigv = np.linalg.eigh(ovp)
  assert (np.all(eigs>=0))

  X = scipy.linalg.sqrtm(ovp)
  owfc = np.linalg.solve(X.T, wfc) 
  
  wfc = np.array(wfc) * scalef
  gkspace = { 'xk': xk, 'igwx': igwx, 'mill': mill, 'bg': bg, 'gamma_only': gamma_only }
  return gkspace, { 'wfc': owfc, 'npol': npol, 'nbnd': nbnd, 'ispin': ispin }


def calc_ylmg(k_plus_G, q):
    # cubic harmonics: build the angular part, no spin orbit
    kGx = np.zeros_like(q)
    kGy = np.zeros_like(q)
    kGz = np.zeros_like(q)
    for ig in range(len(q)):
      if abs(q[ig]) > 1e-6:
        kGx[ig] = k_plus_G[ig,0] / q[ig]
        kGy[ig] = k_plus_G[ig,1] / q[ig]
        kGz[ig] = k_plus_G[ig,2] / q[ig]
    
    lmax = 3
    ylmg = np.zeros((len(q), (lmax+1)*(lmax+1)))

    # normalization factors for l=0,1,2,3
    n0 = np.sqrt(1.0/(4.0*np.pi))
    n1 = np.sqrt(3.0/(4.0*np.pi))
    n2 = np.sqrt(15.0/(4.0*np.pi))
    n3 = np.sqrt(105.0/(4.0*np.pi))

    # l = 0
    ylmg[:,0] = n0 * np.ones_like(q)

    # l = 1
    ylmg[:,1] = n1 * kGz
    ylmg[:,2] = -n1 * kGx         # note the '-' sign, due to the (lack of) Condon–Shortley phase
    ylmg[:,3] = -n1 * kGy         # the same in the l=2 and l=3 blocks, when |m| is odd

    # l = 2
    ylmg[:,4] = n2 * (3*kGz*kGz - 1)/(2*np.sqrt(3))
    ylmg[:,5] = -n2 * kGz*kGx
    ylmg[:,6] = -n2 * kGy*kGz
    ylmg[:,7] = n2 * (kGx*kGx - kGy*kGy) / 2.0
    ylmg[:,8] = n2 * kGx * kGy

    # l = 3
    ylmg[:,9] = n3 * kGz * (2*kGz*kGz - 3*kGx*kGx - 3*kGy*kGy) / (2.0*np.sqrt(15.0))
    ylmg[:,10]= -n3 * kGx * (4*kGz*kGz - kGx*kGx - kGy*kGy) / (2.0*np.sqrt(10.0))
    ylmg[:,11]= -n3 * kGy * (4*kGz*kGz - kGx*kGx - kGy*kGy) / (2.0*np.sqrt(10.0))
    ylmg[:,12]= n3 * kGz * (kGx*kGx - kGy*kGy) / 2.0
    ylmg[:,13]= n3 * kGx * kGy * kGz
    ylmg[:,14]= -n3 * kGx * (kGx*kGx - 3*kGy*kGy) / (2.0*np.sqrt(6.0))
    ylmg[:,15]= -n3 * kGy * (3*kGx*kGx - kGy*kGy) / (2.0*np.sqrt(6.0))

    return ylmg


def calc_ylmg_complex_0(ylmg):
    # complex spherical harmonics
    ylmgc = np.zeros_like(ylmg, np.complex128)

    sqrt2 = np.sqrt(2.0)

    # l = 0
    ylmgc[:,0] = ylmg[:,0]

    # l = 1
    ylmgc[:,1] = ylmg[:,1]
    #ylmgc[:,2] = -(ylmg[:,2] + 1j*ylmg[:,3])/sqrt2   # m=1
    #ylmgc[:,3] =  (ylmg[:,2] - 1j*ylmg[:,3])/sqrt2   # m=-1
    ylmgc[:,2] =  (ylmg[:,2] + 1j*ylmg[:,3])/sqrt2   # m=1     # because QE cub. harm. have opposite sign
    ylmgc[:,3] = -(ylmg[:,2] - 1j*ylmg[:,3])/sqrt2   # m=-1

    # l = 2
    ylmgc[:,4] = ylmg[:,4]
    #ylmgc[:,5] = -(ylmg[:,5] + 1j*ylmg[:,6])/sqrt2   # m=1
    #ylmgc[:,6] =  (ylmg[:,5] - 1j*ylmg[:,6])/sqrt2   # m=-1
    ylmgc[:,5] =  (ylmg[:,5] + 1j*ylmg[:,6])/sqrt2   # m=1     # because QE cub. harm. have opposite sign
    ylmgc[:,6] = -(ylmg[:,5] - 1j*ylmg[:,6])/sqrt2   # m=-1
    ylmgc[:,7] = +(ylmg[:,7] + 1j*ylmg[:,8])/sqrt2   # m=2
    ylmgc[:,8] =  (ylmg[:,7] - 1j*ylmg[:,8])/sqrt2   # m=-2
    
    # l = 3
    ylmgc[:,9] = ylmg[:,9]
    #ylmgc[:,10] = -(ylmg[:,10] + 1j*ylmg[:,11])/sqrt2   # m=1
    #ylmgc[:,11] =  (ylmg[:,10] - 1j*ylmg[:,11])/sqrt2   # m=-1
    ylmgc[:,10] =  (ylmg[:,10] + 1j*ylmg[:,11])/sqrt2   # m=1     # because QE cub. harm. have opposite sign
    ylmgc[:,11] = -(ylmg[:,10] - 1j*ylmg[:,11])/sqrt2   # m=-1
    ylmgc[:,12] = +(ylmg[:,12] + 1j*ylmg[:,13])/sqrt2   # m=2
    ylmgc[:,13] =  (ylmg[:,12] - 1j*ylmg[:,13])/sqrt2   # m=-2
    #ylmgc[:,14] = -(ylmg[:,14] + 1j*ylmg[:,15])/sqrt2   # m=3
    #ylmgc[:,15] =  (ylmg[:,14] - 1j*ylmg[:,15])/sqrt2   # m=-3
    ylmgc[:,14] =  (ylmg[:,14] + 1j*ylmg[:,15])/sqrt2   # m=3     # because QE cub. harm. have opposite sign
    ylmgc[:,15] = -(ylmg[:,14] - 1j*ylmg[:,15])/sqrt2   # m=-3

    return ylmgc


def calc_ylmg_so(ylmgc):
    # spinor spherical harmonics
    npw = ylmgc.shape[0]
    nylm = ylmgc.shape[1]
    ylmgso = np.zeros((2*npw,2*nylm), np.complex128)
    sqrt = np.sqrt

    # generated automatically by cb.py
    #l=0, j=0.5 m_j=-0.5 upper=sqrt(0)*Y(0,-1) 	 lower=sqrt(1)*Y(0,0)
    ylmgso[:npw,0]=0.0; ylmgso[npw:,0]=sqrt(1)*ylmgc[:npw,0]; 

    #l=0, j=0.5 m_j= 0.5 upper=sqrt(1)*Y(0,0) 	 lower=sqrt(0)*Y(0,1)
    ylmgso[:npw,1]=sqrt(1)*ylmgc[:npw,0]; ylmgso[npw:,1]=0.0; 

    #l=1, j=0.5 m_j=-0.5 upper=sqrt(2/3)*Y(1,-1) 	 lower=-sqrt(1/3)*Y(1,0))
    ylmgso[:npw,2]=sqrt(2/3)*ylmgc[:npw,3]; ylmgso[npw:,2]=-sqrt(1/3)*ylmgc[:npw,1]; 

    #l=1, j=0.5 m_j= 0.5 upper=sqrt(1/3)*Y(1,0) 	 lower=-sqrt(2/3)*Y(1,1))
    ylmgso[:npw,3]=sqrt(1/3)*ylmgc[:npw,1]; ylmgso[npw:,3]=-sqrt(2/3)*ylmgc[:npw,2]; 

    #l=1, j=1.5 m_j=-1.5 upper=sqrt(0)*Y(1,-2) 	 lower=sqrt(1)*Y(1,-1)
    ylmgso[:npw,4]=0.0; ylmgso[npw:,4]=sqrt(1)*ylmgc[:npw,3]; 

    #l=1, j=1.5 m_j=-0.5 upper=sqrt(1/3)*Y(1,-1) 	 lower=sqrt(2/3)*Y(1,0)
    ylmgso[:npw,5]=sqrt(1/3)*ylmgc[:npw,3]; ylmgso[npw:,5]=sqrt(2/3)*ylmgc[:npw,1]; 

    #l=1, j=1.5 m_j= 0.5 upper=sqrt(2/3)*Y(1,0) 	 lower=sqrt(1/3)*Y(1,1)
    ylmgso[:npw,6]=sqrt(2/3)*ylmgc[:npw,1]; ylmgso[npw:,6]=sqrt(1/3)*ylmgc[:npw,2]; 

    #l=1, j=1.5 m_j= 1.5 upper=sqrt(1)*Y(1,1) 	 lower=sqrt(0)*Y(1,2)
    ylmgso[:npw,7]=sqrt(1)*ylmgc[:npw,2]; ylmgso[npw:,7]=0.0; 

    #l=2, j=1.5 m_j=-1.5 upper=sqrt(4/5)*Y(2,-2) 	 lower=-sqrt(1/5)*Y(2,-1))
    ylmgso[:npw,8]=sqrt(4/5)*ylmgc[:npw,8]; ylmgso[npw:,8]=-sqrt(1/5)*ylmgc[:npw,6]; 

    #l=2, j=1.5 m_j=-0.5 upper=sqrt(3/5)*Y(2,-1) 	 lower=-sqrt(2/5)*Y(2,0))
    ylmgso[:npw,9]=sqrt(3/5)*ylmgc[:npw,6]; ylmgso[npw:,9]=-sqrt(2/5)*ylmgc[:npw,4]; 

    #l=2, j=1.5 m_j= 0.5 upper=sqrt(2/5)*Y(2,0) 	 lower=-sqrt(3/5)*Y(2,1))
    ylmgso[:npw,10]=sqrt(2/5)*ylmgc[:npw,4]; ylmgso[npw:,10]=-sqrt(3/5)*ylmgc[:npw,5]; 

    #l=2, j=1.5 m_j= 1.5 upper=sqrt(1/5)*Y(2,1) 	 lower=-sqrt(4/5)*Y(2,2))
    ylmgso[:npw,11]=sqrt(1/5)*ylmgc[:npw,5]; ylmgso[npw:,11]=-sqrt(4/5)*ylmgc[:npw,7]; 

    #l=2, j=2.5 m_j=-2.5 upper=sqrt(0)*Y(2,-3) 	 lower=sqrt(1)*Y(2,-2)
    ylmgso[:npw,12]=0.0; ylmgso[npw:,12]=sqrt(1)*ylmgc[:npw,8]; 

    #l=2, j=2.5 m_j=-1.5 upper=sqrt(1/5)*Y(2,-2) 	 lower=sqrt(4/5)*Y(2,-1)
    ylmgso[:npw,13]=sqrt(1/5)*ylmgc[:npw,8]; ylmgso[npw:,13]=sqrt(4/5)*ylmgc[:npw,6]; 

    #l=2, j=2.5 m_j=-0.5 upper=sqrt(2/5)*Y(2,-1) 	 lower=sqrt(3/5)*Y(2,0)
    ylmgso[:npw,14]=sqrt(2/5)*ylmgc[:npw,6]; ylmgso[npw:,14]=sqrt(3/5)*ylmgc[:npw,4]; 

    #l=2, j=2.5 m_j= 0.5 upper=sqrt(3/5)*Y(2,0) 	 lower=sqrt(2/5)*Y(2,1)
    ylmgso[:npw,15]=sqrt(3/5)*ylmgc[:npw,4]; ylmgso[npw:,15]=sqrt(2/5)*ylmgc[:npw,5]; 

    #l=2, j=2.5 m_j= 1.5 upper=sqrt(4/5)*Y(2,1) 	 lower=sqrt(1/5)*Y(2,2)
    ylmgso[:npw,16]=sqrt(4/5)*ylmgc[:npw,5]; ylmgso[npw:,16]=sqrt(1/5)*ylmgc[:npw,7]; 

    #l=2, j=2.5 m_j= 2.5 upper=sqrt(1)*Y(2,2) 	 lower=sqrt(0)*Y(2,3)
    ylmgso[:npw,17]=sqrt(1)*ylmgc[:npw,7]; ylmgso[npw:,17]=0.0; 

    #l=3, j=2.5 m_j=-2.5 upper=sqrt(6/7)*Y(3,-3) 	 lower=-sqrt(1/7)*Y(3,-2))
    ylmgso[:npw,18]=sqrt(6/7)*ylmgc[:npw,15]; ylmgso[npw:,18]=-sqrt(1/7)*ylmgc[:npw,13]; 

    #l=3, j=2.5 m_j=-1.5 upper=sqrt(5/7)*Y(3,-2) 	 lower=-sqrt(2/7)*Y(3,-1))
    ylmgso[:npw,19]=sqrt(5/7)*ylmgc[:npw,13]; ylmgso[npw:,19]=-sqrt(2/7)*ylmgc[:npw,11]; 

    #l=3, j=2.5 m_j=-0.5 upper=sqrt(4/7)*Y(3,-1) 	 lower=-sqrt(3/7)*Y(3,0))
    ylmgso[:npw,20]=sqrt(4/7)*ylmgc[:npw,11]; ylmgso[npw:,20]=-sqrt(3/7)*ylmgc[:npw,9]; 

    #l=3, j=2.5 m_j= 0.5 upper=sqrt(3/7)*Y(3,0) 	 lower=-sqrt(4/7)*Y(3,1))
    ylmgso[:npw,21]=sqrt(3/7)*ylmgc[:npw,9]; ylmgso[npw:,21]=-sqrt(4/7)*ylmgc[:npw,10]; 

    #l=3, j=2.5 m_j= 1.5 upper=sqrt(2/7)*Y(3,1) 	 lower=-sqrt(5/7)*Y(3,2))
    ylmgso[:npw,22]=sqrt(2/7)*ylmgc[:npw,10]; ylmgso[npw:,22]=-sqrt(5/7)*ylmgc[:npw,12]; 

    #l=3, j=2.5 m_j= 2.5 upper=sqrt(1/7)*Y(3,2) 	 lower=-sqrt(6/7)*Y(3,3))
    ylmgso[:npw,23]=sqrt(1/7)*ylmgc[:npw,12]; ylmgso[npw:,23]=-sqrt(6/7)*ylmgc[:npw,14]; 

    #l=3, j=3.5 m_j=-3.5 upper=sqrt(0)*Y(3,-4) 	 lower=sqrt(1)*Y(3,-3)
    ylmgso[:npw,24]=0.0; ylmgso[npw:,24]=sqrt(1)*ylmgc[:npw,15]; 

    #l=3, j=3.5 m_j=-2.5 upper=sqrt(1/7)*Y(3,-3) 	 lower=sqrt(6/7)*Y(3,-2)
    ylmgso[:npw,25]=sqrt(1/7)*ylmgc[:npw,15]; ylmgso[npw:,25]=sqrt(6/7)*ylmgc[:npw,13]; 

    #l=3, j=3.5 m_j=-1.5 upper=sqrt(2/7)*Y(3,-2) 	 lower=sqrt(5/7)*Y(3,-1)
    ylmgso[:npw,26]=sqrt(2/7)*ylmgc[:npw,13]; ylmgso[npw:,26]=sqrt(5/7)*ylmgc[:npw,11]; 

    #l=3, j=3.5 m_j=-0.5 upper=sqrt(3/7)*Y(3,-1) 	 lower=sqrt(4/7)*Y(3,0)
    ylmgso[:npw,27]=sqrt(3/7)*ylmgc[:npw,11]; ylmgso[npw:,27]=sqrt(4/7)*ylmgc[:npw,9]; 

    #l=3, j=3.5 m_j= 0.5 upper=sqrt(4/7)*Y(3,0) 	 lower=sqrt(3/7)*Y(3,1)
    ylmgso[:npw,28]=sqrt(4/7)*ylmgc[:npw,9]; ylmgso[npw:,28]=sqrt(3/7)*ylmgc[:npw,10]; 

    #l=3, j=3.5 m_j= 1.5 upper=sqrt(5/7)*Y(3,1) 	 lower=sqrt(2/7)*Y(3,2)
    ylmgso[:npw,29]=sqrt(5/7)*ylmgc[:npw,10]; ylmgso[npw:,29]=sqrt(2/7)*ylmgc[:npw,12]; 

    #l=3, j=3.5 m_j= 2.5 upper=sqrt(6/7)*Y(3,2) 	 lower=sqrt(1/7)*Y(3,3)
    ylmgso[:npw,30]=sqrt(6/7)*ylmgc[:npw,12]; ylmgso[npw:,30]=sqrt(1/7)*ylmgc[:npw,14]; 

    #l=3, j=3.5 m_j= 3.5 upper=sqrt(1)*Y(3,3) 	 lower=sqrt(0)*Y(3,4)
    ylmgso[:npw,31]=sqrt(1)*ylmgc[:npw,14]; ylmgso[npw:,31]=0.0; 

    return ylmgso    



def calc_atwfc_k(basis, gkspace, dftSO=False):
  # construct atomic wfc at k
  atwfc_k = []
  natwfc = len(basis)
  
  xk, igwx, mill, bg, gamma_only = [gkspace[s] for s in ('xk', 'igwx', 'mill', 'bg', 'gamma_only')]
  #print('xk=', xk)
  
  # build k+G vectors
  hkl = mill.T
  k_plus_G = np.dot(hkl, bg.T)
  for ig in range(igwx):
      k_plus_G[ig,:] += xk

  # pre-calculate spherical harmonics
  q = np.linalg.norm(k_plus_G, axis=1)
  ylmg = calc_ylmg(k_plus_G, q)
  if dftSO:
      ylmgc = calc_ylmg_complex_0(ylmg)
      ylmgso = calc_ylmg_so(ylmgc)
 
  # loop over atoms 
  for i in range(natwfc):
    # 1. build the structure factor
    strf = np.zeros((igwx,), dtype=complex)
    tau = basis[i]['tau']
    k_plus_G_dot_tau = np.dot(k_plus_G, tau)
    strf = np.exp(-1j*k_plus_G_dot_tau)
    
    # 2. build the form factor
    qmesh, wfc_g = basis[i]['qmesh'], basis[i]['wfc_g']
    #fact = InterpolatedUnivariateSpline(qmesh, wfc_g)(q)
    fact = scipy.interpolate.interp1d(qmesh, wfc_g, kind='linear')(q)
    
    # 3. build the angular part
    l, m = basis[i]['l'], basis[i]['m']
    if l > 3: raise NotImplementedError('l>3 not implemented yet')
    lm = l*l + (m-1)
    if dftSO:
        jm = basis[i]['jm']

    # 4. final
    if not dftSO:
        atwfc = strf * fact * ylmg[:,lm] * (1.0j)**l
    else:
        atwfc = np.hstack((strf,strf)) * np.hstack((fact,fact)) * ylmgso[:,jm] * (1.0j)**l
 
    atwfc_k.append(atwfc)
    
  return np.array(atwfc_k)



def ortho_atwfc_k(atwfc_k):
  # orthonormalize atwfcs
  natwfc = atwfc_k.shape[0]
  ovp = np.zeros((natwfc, natwfc), dtype=np.complex128)
  
  for i in range(natwfc):
    for j in range(natwfc):
      ovp[i,j] = np.dot(np.conj(atwfc_k[i]), atwfc_k[j])
      
      # check that eigenvalues are positive
  eigs, eigv = np.linalg.eigh(ovp)
  assert (np.all(eigs>=0))
  
  # orthogonalize
  if True:
    X = scipy.linalg.sqrtm(ovp)
    oatwfc_k = np.linalg.solve(X.T, atwfc_k)
  else:
    eigs = 1.0/np.sqrt(eigs)
    X = np.dot(np.conj(eigv), np.dot(np.diag(eigs), eigv.T))
    oatwfc_k = np.dot(X, atwfc_k)
    
    # check ortonormalization
  oovp = np.zeros((natwfc, natwfc), dtype=np.complex128)
  for i in range(natwfc):
    for j in range(natwfc):
      oovp[i,j] = np.dot(np.conj(oatwfc_k[i]), oatwfc_k[j])

  # np.save("oovp",oovp)
  diff = np.linalg.norm(oovp - np.eye(natwfc))
  if np.abs(diff) > 1e-4:
    raise RuntimeError('ortogonalization failed')
    
  return oatwfc_k


def calc_proj_k(data_controller, basis, ik, ispin):
  arry, attr = data_controller.data_dicts()
  if attr['dft'] == 'QE':
    gkspace, wfc = read_QE_wfc(data_controller, ik, ispin)
  else:
    read_WAVECAR_header(data_controller)
    gkspace, wfc = read_VASP_wfc(data_controller, ik, ispin)
  atwfc_k = calc_atwfc_k(basis, gkspace, attr['dftSO'])
  oatwfc_k = ortho_atwfc_k(atwfc_k)
  proj_k = np.dot(np.conj(oatwfc_k), wfc['wfc'].T)
  return (proj_k.T)


def calc_gkspace(data_controller,ik,gamma_only=False):
  arry, attr = data_controller.data_dicts()
  # calculate sphere of Miller indeces for k + G
  gcutm = attr['ecutrho'] / (2*np.pi/attr['alat'])**2
  at = arry['a_vectors']
  nx = 2*int(np.sqrt(gcutm)*np.sqrt(at[0,0]**2 + at[1,0]**2 + at[2,0]**2)) + 1
  ny = 2*int(np.sqrt(gcutm)*np.sqrt(at[0,1]**2 + at[1,1]**2 + at[2,1]**2)) + 1
  nz = 2*int(np.sqrt(gcutm)*np.sqrt(at[0,2]**2 + at[1,2]**2 + at[2,2]**2)) + 1
  nx = int((nx+1)/2)
  ny = int((ny+1)/2)
  nz = int((nz+1)/2)
  # sphere of Miller indeces 
  mill_g = []
  for i in range(-nx, nx+1):
    for j in range(-ny, ny+1):
      for k in range(-nz, nz+1):
        k_plus_G = np.array([i,j,k]) @ arry['b_vectors']
        if np.linalg.norm(k_plus_G)**2 <= attr['ecutrho']/(2*np.pi/attr['alat'])**2:
          mill_g.append(np.array([i,j,k]))
  mill_g = np.swapaxes(np.array(mill_g),1,0)
  igwx_g = mill_g.shape[1]
  
  mill = []
  for ig in range(igwx_g):
    k_plus_G = mill_g[:,ig]@arry['b_vectors'] + arry['kgrid'][:,ik]
    if np.linalg.norm(k_plus_G)**2 <= attr['ecutwfc']/(2*np.pi/attr['alat'])**2:
      mill.append(mill_g[:,ig])
  mill = np.swapaxes(np.array(mill),1,0)
  igwx = mill.shape[1]
  
  xk = arry['kgrid'][:,ik] * 2*np.pi/attr['alat']
  
  bg = arry['b_vectors'].T*2*np.pi/attr['alat']
  
  names = ['xk','igwx','mill','bg','gamma_only']
  arrays = [xk,igwx,mill,bg,gamma_only]
  gkspace = dict(zip(names,arrays))
  arry['gkspace'] = gkspace
  
  return(gkspace)

## Added functions for VASP interface - ZH
# Credit to Qijing Zheng, https://github.com/QijingZheng/VaspBandUnfolding
def read_WAVECAR_header(data_controller):
  arry, attr = data_controller.data_dicts()
  try:
    wfcfile = open(os.path.join(attr['fpath'], 'WAVECAR'), 'rb')
  except:
    print('WAVECAR was not found.')
  # go to the start of the file and read the first record
  wfcfile.seek(0)
  recl, nspin, rtag = np.array(
    np.fromfile(wfcfile, dtype=np.float64, count=3),
    dtype=np.int64
  )
  attr['recl'] = recl
  if nspin != attr['nspin']:
    raise Exception("NSPIN in vasprun.xml and WAVECAR differ")
  if rtag == 45200:
    attr['wfc_prec'] = np.complex64
  elif rtag == 45210:
    attr['wfc_prec'] = np.complex128
  else:
    raise Exception("Invalid rtag in WAVECAR")

  wfcfile.seek(recl)
  dump = np.fromfile(wfcfile, dtype=np.float64, count=12)
  nkpnts = int(dump[0])
  nbnds = int(dump[1])
  encut = dump[2] # in unit of eV
  Acell = dump[3:].reshape((3, 3))/AUTOA # in unit of Bohr

  # Check vasprun.xml and WAVECAR have the same parameters
  if nkpnts != attr['nkpnts']:
    raise Exception("Incorrect # of k-points. Restart from beginning or check if vasprun.xml and WAVECAR come from the same run")
  if nbnds != attr['nbnds']:
    raise Exception("# of bands in vasprun.xml and WAVECAR differ")
  if abs(encut-attr['ecutwfc']*13.605826) > 1e-6:
    raise Exception("ENCUT in vasprun.xml and WAVECAR differ")
  if np.linalg.norm(arry['a_vectors']-Acell) > 1e-6:
    raise Exception("Lattice in vasprun.xml and WAVECAR differ")

  Anorm = np.linalg.norm(Acell, axis=1)
  CUTOF = np.ceil(np.sqrt(attr['ecutwfc']) / (2 * np.pi / Anorm))
  ngrid = np.array(2 * CUTOF + 1, dtype=int)
  arry['ngrid'] = ngrid

  nplws = np.zeros(nkpnts, dtype=int)

  for ispin in range(nspin):
    for ik in range(nkpnts):
      rec = 2 + ispin * nkpnts * (nbnds + 1) + ik * (nbnds + 1)
      wfcfile.seek(rec * recl)
      dump = np.fromfile(wfcfile, dtype=np.float64, count=4 + 3 * nbnds)
      if ispin == 0:
        nplws[ik] = int(dump[0])

  arry['nplws'] = nplws
 

def readBandCoeff(data_controller, ispin=0, ikpt=0, iband=0):
  # Read the planewave coefficients
  arry, attr = data_controller.data_dicts()
  rec = 3 + ispin * attr['nkpnts'] * (attr['nbnds'] + 1) + ikpt * (attr['nbnds'] + 1) + iband
  wfcfile = open(os.path.join(attr['fpath'], 'WAVECAR'), 'rb')
  wfcfile.seek(rec * attr['recl'])
  dump = np.fromfile(wfcfile, dtype=attr['wfc_prec'], count=arry['nplws'][ikpt])
  return np.asarray(dump, dtype=attr['wfc_prec'])

def read_VASP_wfc(data_controller, ik, ispin):
  arry, attr = data_controller.data_dicts()
  igwx = arry['nplws'][ik] // 2 if attr['dftSO'] else arry['nplws'][ik]
  nbnd = attr['nbnds']
  bg = 2*np.pi*arry['b_vectors'].T
  xk = 2*np.pi*arry['kpnts'][ik, :]
  mill = calc_gvec_VASP(data_controller, ik).T

  wfc = []
  for ibnd in range(nbnd):
    wfc.append(readBandCoeff(data_controller, ispin, ik, ibnd))
  # compute overlap
  ovp = np.zeros((nbnd, nbnd), dtype=complex)
  for n in range(nbnd):
    for m in range(nbnd):
      ovp[n, m] = np.sum(np.conj(wfc[n]).dot(wfc[m]))
  eigs, eigv = np.linalg.eigh(ovp)
  assert (np.all(eigs >= 0))

  X = scipy.linalg.sqrtm(ovp)
  owfc = np.linalg.solve(X.T, wfc)

  gkspace = {'xk': xk, 'igwx': igwx, 'mill': mill, 'bg': bg, 'gamma_only': False}  # Gamma-only not supported yet
  return gkspace, {'wfc': owfc, 'npol': attr['nspin'], 'nbnd': nbnd, 'ispin': ispin}

def calc_gvec_VASP(data_controller, ikpt=0):

  arry, attr = data_controller.data_dicts()
  kvec = arry['kpnts'][ikpt]
  nplw = arry['nplws'][ikpt]
  ngrid = arry['ngrid']
  lgam = False  # Gamma-only calculations not supported

  fx, fy, fz = [np.arange(n, dtype=int) for n in ngrid]
  fx[ngrid[0] // 2 + 1:] -= ngrid[0]
  fy[ngrid[1] // 2 + 1:] -= ngrid[1]
  fz[ngrid[2] // 2 + 1:] -= ngrid[2]

  # In meshgrid, fx run the fastest, fz the slowest
  gz, gy, gx = np.array(
    np.meshgrid(fz, fy, fx, indexing='ij')
  ).reshape((3, -1))
  kgrid = np.array([gx, gy, gz], dtype=float).T

  # Kinetic_Energy = (G + k)**2 / 2
  kenergy = (2 * np.pi * np.linalg.norm(
    np.dot(kgrid, arry['b_vectors']) + kvec, axis=1
  )) ** 2
  # find Gvectors where (G + k)**2 / 2 < ENCUT
  Gvec = kgrid[np.where(kenergy < attr['ecutwfc'])[0]]

  # Check if the calculated number of planewaves and the one recorded in the
  # WAVECAR are equal
  if Gvec.shape[0] != nplw:
    if Gvec.shape[0] * 2 == nplw:
      if not attr['dftSO']:
        raise ValueError('''
                No. of planewaves NOT consistent: noncollinear calculation
                ''')
    elif Gvec.shape[0] == 2 * nplw - 1:
      if not lgam:
        raise ValueError('''
                No. of planewaves NOT consistent: Gamma-only calculation. 
                ''')
    else:
      raise ValueError('''
            No. of planewaves NOT consistent:
                this code -> %d
                from VASP -> %d
            ''' % (Gvec.shape[0],
                   nplw // 2 if attr['dftSO'] else nplw)
                  )

  return np.asarray(Gvec, dtype=int)



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
#F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli, Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import numpy as np

class PAOFLOW:

  data_controller = None

  comm = rank = size = None

  start_time = reset_time = None

  # Overestimate factor for guessing memory requirements
  gb_fudge_factor = 4.

  # Function container for ErrorHandler's method
  report_exception = None


  def print_data_keys ( self ):
    '''
      Print's out the keys do the data_controller dictionaries, "arrays" and "attributes".
      Each stores the respective data required for the various calculation PAOFLOW can perform.

      Arguments:
          None

      Returns:
          None
    '''
    if self.rank == 0:
      self.data_controller.print_data()
    self.comm.Barrier()



  def __init__ ( self, workpath='./', outputdir='output', inputfile=None, savedir=None, model=None, npool=1, smearing='gauss', acbn0=False, verbose=False, restart=False, dft='QE'):
    '''
    Initialize the PAOFLOW class, either with a save directory with required QE output or with an xml inputfile
    Arguments:
        workpath (str): Path to the working directory
        outputdir (str): Name of the output directory (created in the working directory path)
        inputfile (str): (optional) Name of the xml inputfile
        savedir (str): QE .save directory
        model (dict): Dictionary with 'label' key and parameters to build Hamiltonian from TB model
        npool (int): The number of pools to use. Increasing npool may reduce memory requirements.
        smearing (str): Smearing type (None, m-p, gauss)
        acbn0 (bool): If True the Hamiltonian will be Orthogonalized after construction
        verbose (bool): False supresses debugging output
        restart (bool): True if the run is being restarted from a .json data dump.
        dft (str): 'QE' or 'VASP'
    Returns:
        None
    '''
    from time import time
    from mpi4py import MPI
    from .defs.header import header
    from .DataController import DataController

    #-------------------------------
    # Initialize Parallel Execution
    #-------------------------------
    self.comm = MPI.COMM_WORLD
    self.rank = self.comm.Get_rank()
    self.size = self.comm.Get_size()

    #-----------------
    # Print Header
    # Initialize Time
    #--------------
    if self.rank == 0:
      header()
      self.start_time = self.reset_time = time()

    # Initialize Data Controller
    self.data_controller = DataController(workpath, outputdir, inputfile, model, savedir, npool, smearing, acbn0, verbose, restart, dft)

    self.report_exception = self.data_controller.report_exception

    if not restart:
      # Data Attributes
      attr = self.data_controller.data_attributes

      # Check for CUDA FFT Libraries
      ## CUDA not yet supported in PAOFLOW_CLASS
      attr['use_cuda'] = False
      attr['scipyfft'] = True
      if attr['use_cuda']:
        attr['scipyfft'] = False
      if self.rank == 0 and attr['verbose']:
        if attr['use_cuda']:
          print('CUDA will perform FFTs on %d GPUs'%1)
        else:
          print('SciPy will perform FFTs')

    # Report execution information
    if self.rank == 0:
      if restart:
        print('Run starting from Restart data.')
      else:
        if self.size == 1:
          print('Serial execution')
        else:
          print('Parallel execution on %d processors and %d pool'%(self.size,attr['npool']) + ('' if attr['npool']==1 else 's'))

    # Do memory checks
    if model is None and not restart and self.rank == 0:
      gbyte = self.memory_check()
      print('Estimated maximum array size: %.2f GBytes\n' %(gbyte))

    self.report_module_time('Initialization')



  def report_module_time ( self, mname ):
    from time import time

    self.comm.Barrier()

    if self.rank == 0:

      # White spacing between module name and reported time
      spaces = 35
      lmn = len(mname)
      if len(mname) > spaces:
        print('DEBUG: Please use a shorter module tag.')
        self.comm.Abort()

      # Format string and print
      lms = spaces-lmn
      dt = time() - self.reset_time
      print('%s in: %s %8.3f sec'%(mname,lms*' ',dt), flush=True)
      self.reset_time = time()



  def restart_dump ( self, fname_prefix='paoflow_dump' ):
    '''
      Saves the necessary information to restart a PAOFLOW run from any step in calculation.

      Arguments:
          fname_prefix (str): Prefix of the filenames which will be written. Files are written to the directory housing the python script which instantiates PAOFLOW, unless otherwise specified in this argument.

      Returns:
          None
    '''
    from pickle import dump,HIGHEST_PROTOCOL

    fname = fname_prefix + '_%d'%self.rank + '.json'

    arry,attr = self.data_controller.data_dicts()
    with open(fname, 'wb') as f:
      dump([arry,attr], f, HIGHEST_PROTOCOL)

    self.report_module_time('Restart DUMP')



  def restart_load ( self, fname_prefix='paoflow_dump' ):
    '''
      Loads the previously dumped save files and populates the DataController with said data.

      Arguments:
          fname_prefix (str): Prefix of the filenames which will be written. Files are written to the directory housing the python script which instantiates PAOFLOW, unless otherwise specified in this argument.

      Returns:
          None
    '''
    from os.path import exists
    from pickle import load

    fname = fname_prefix + '_%d'%self.rank + '.json'
    if not exists(fname):
      print('Restart file named %s does not exist.'%fname)
      raise OSError('File: %s not found.'%fname)

    arry,attr = None,None
    with open(fname, 'rb') as f:
      arry,attr = load(f)

    if self.size != attr['mpisize']:
      print('Restarted runs must use the same number of cores as the original run.')
      raise ValueError('Number of processors does not match that of the previous run.')

    self.data_controller.data_arrays = arry
    self.data_controller.data_attributes = attr

    self.report_module_time('Restart LOAD')



  def memory_check ( self ):
    '''
    Estimate PAOFLOW's memory requirements with a "fudge factor" defined above

    Arguments:
        None

    Returns:
        gbyte (float): Estimated number of Gigabytes required in memory for PAOFLOW to run.
    '''
    dxdydz = 3
    B_to_GB = 1.E-9
    ff = self.gb_fudge_factor
    bytes_per_complex = 128//8
    arry,attr = self.data_controller.data_dicts()
    nd1,nd2,nd3 = attr['nk1'],attr['nk2'],attr['nk3']
    spins,num_wave_functions = attr['nspin'],attr['nawf']
    return num_wave_functions**2 * (nd1*nd2*nd3) * spins * dxdydz * bytes_per_complex * ff * B_to_GB



  def finish_execution ( self ):
    '''
    Finish the PAOFLOW execution. Print out run time and maximum memory usage

    Arguments:
        None

    Returns:
        None
    '''
    import resource
    from time import time
    from mpi4py import MPI

    if self.rank == 0:
      tt = time() - self.start_time
      print('Total CPU time =%s%8.3f sec'%(25*' ',tt))

    verbose = self.data_controller.data_attributes['verbose']

    if verbose:

      # Add up memory usage from each core
      mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
      mem = np.array([mem], dtype=float)
      mem0 = np.zeros(1, dtype=float) if self.rank==0 else None
      self.comm.Reduce(mem, mem0, op=MPI.SUM, root=0)

      if self.rank == 0:
        print("Memory usage on rank 0:  %6.4f GB"%(mem[0]/1024.0**2))
        print("Maximum concurrent memory usage:  %6.4f GB"%(mem0[0]/1024.0**2))



  def projections ( self, internal=False, basispath=None, configuration=None ):
    '''
    Calculate the projections on the atomic basis provided by the pseudopotential or 
    on the all-electron internal basis sets.
    Replaces projwfc.
    '''

    from .defs.do_atwfc_proj import build_pswfc_basis_all
    from .defs.do_atwfc_proj import build_aewfc_basis
    from .defs.do_atwfc_proj import calc_proj_k
    from .defs.communication import load_balancing, gather_array
    
    arry,attr = self.data_controller.data_dicts()

    if basispath is not None:
      attr['basispath'] = basispath
    if configuration is not None:
      arry['configuration'] = configuration

    # Always use internal basis if VASP
    if internal or attr['dft']=='VASP':
      basis,arry['shells'] = build_aewfc_basis(self.data_controller)
    else:
      basis,arry['shells'] = build_pswfc_basis_all(self.data_controller)

    arry['Dnm'] = np.empty((attr['nawf'],attr['nawf'],3))
    for i in range(3):
      for n in range(attr['nawf']):
        for m in range(attr['nawf']):
          arry['Dnm'][n,m,i] = basis[n]['tau'][i] - basis[m]['tau'][i]

    nkpnts = len(arry['kpnts'])
    nbnds = attr['nbnds']
    nspin = attr['nspin']
    natwfc = len(basis)
    attr['nawf'] = natwfc
    
    ini_ik,end_ik = load_balancing(self.size, self.rank,nkpnts)
    Unewaux = np.zeros((end_ik-ini_ik,nbnds,natwfc,nspin), dtype=complex)
    for ispin in range(nspin):
      for ik in range(ini_ik,end_ik):
        Unewaux[ik-ini_ik,:,:,ispin] = calc_proj_k(self.data_controller, basis, ik, ispin)
    
    Unew = np.zeros((nkpnts,nbnds,natwfc,nspin), dtype=complex) if self.rank == 0 else None
    gather_array(Unew,Unewaux)
    if self.rank == 0: Unew = np.moveaxis(Unew,0,2)
    Unew = self.comm.bcast(Unew, root=0)
    
    arry['U'] = Unew
    arry['basis'] = basis
    
    self.report_module_time('Projections')



  def read_atomic_proj_QE ( self ):
    '''
      Read the wavefunctions and overlaps from atomic-proj.xml, written by Quantum Espresso
      in the .save directory specified in PAOFLOW's constructos. They are saved to the 
      DataController's arrays dictionary with keys 'U' and 'Sks', respectively.
    '''
    from .defs.read_upf import UPF
    from os.path import exists,join
    from .defs.do_atwfc_proj import build_pswfc_basis_all
    from .defs.do_atwfc_proj import build_aewfc_basis

    arry,attr = self.data_controller.data_dicts()
    fpath = attr['fpath']
    if exists(join(fpath,'atomic_proj.xml')):
      from .defs.read_QE_xml import parse_qe_atomic_proj
      parse_qe_atomic_proj(self.data_controller, join(fpath,'atomic_proj.xml'))
    else:
      raise Exception('atomic_proj.xml was not found.\n')

    arry['jchia'] = {}
    arry['shells'] = {}
    for at,pseudo in arry['species']:
      fname = join(attr['fpath'], pseudo)
      if exists(fname):
        upf = UPF(fname)
        arry['shells'][at] = upf.shells
        arry['jchia'][at] = upf.jchia
      else:
        raise Exception('Pseudopotential not found: %s'%fname)

    # Silencing verbose for a moment
    verbose_status=attr['verbose']
    if (verbose_status==True) : attr['verbose']=False
    basis,_ = build_pswfc_basis_all(self.data_controller)     
    # Restoring verbose
    if (verbose_status==True) : attr['verbose']=True

    arry['Dnm'] = np.empty((attr['nawf'],attr['nawf'],3))
    for i in range(3):
      for n in range(attr['nawf']):
        for m in range(attr['nawf']):
          arry['Dnm'][n,m,i] = basis[n]['tau'][i] - basis[m]['tau'][i]

    arry['basis'] = basis


  def projectability ( self, pthr=0.95, shift='auto' ):
    '''
    Calculate the Projectability Matrix to determine how many states need to be shifted

    Arguments:
        pthr (float): The minimum allowed projectability for high projectability states
        shift (str or float): If 'auto' the shift will be automatic. Otherwise the shift will be the value of 'shift'

    Returns:
        None
    '''
    from .defs.do_projectability import do_projectability

    attr = self.data_controller.data_attributes

    if 'pthr' not in attr: attr['pthr'] = pthr
    if 'shift' not in attr: attr['shift'] = shift

    try:
      do_projectability(self.data_controller)
    except Exception as e:
      self.report_exception('projectability')
      if attr['abort_on_exception']:
        raise e

    self.report_module_time('Projectability')
    


  def pao_hamiltonian ( self, shift_type=1, insulator=False, write_binary=False, expand_wedge=True, symmetrize=False, thresh=1.e-6, max_iter=16 ):
    '''
    Construct the Tight Binding Hamiltonian
    Populates DataController with 'HRs', 'Hks' and 'kq_wght'

    Arguments:
        shift_type (int): Shift type [ 0-(PRB 2016), 1-(PRB 2013), 2-No Shift ] 

    Returns:
        None
    
    '''
    from .defs.get_K_grid_fft import get_K_grid_fft
    from .defs.do_build_pao_hamiltonian import do_build_pao_hamiltonian,do_Hks_to_HRs
    from .defs.do_Efermi import E_Fermi

    # Data Attributes and Arrays
    arrays,attr = self.data_controller.data_dicts()

    if insulator: attr['insulator'] = True
    if 'shift_type' not in attr: attr['shift_type'] = shift_type
    if 'write_binary' not in attr: attr['write_binary'] = write_binary

    attr['symm_thresh'] = thresh
    attr['symmetrize'] = symmetrize
    attr['symm_max_iter'] = max_iter
    attr['expand_wedge'] = expand_wedge
    #  Skip open_grid when all k-points are included (no symmetry)
    #  Note expand_wedge is still required for VASP even not using symmetry.
    #  This is because we need find_equiv_k() in paosym to have the correct k-point ordering.

    if attr['symmetrize'] and attr['acbn0']:
      if rank == 0:
        print('WARNING: Non-ortho is currently not supported with pao_sym. Use nosym=.true., noinv=.true.')

    try:
      do_build_pao_hamiltonian(self.data_controller)
    except Exception as e:
      self.report_exception('pao_hamiltonian')
      if attr['abort_on_exception']:
        raise e
    self.report_module_time('Building Hks')

    # Done with U and Sks
    del arrays['U']

    try:
      do_Hks_to_HRs(self.data_controller)

      ### PARALLELIZATION
      self.data_controller.broadcast_single_array('HRs')

      get_K_grid_fft(self.data_controller)
    except Exception as e:
      self.report_exception('pao_hamiltonian')
      if attr['abort_on_exception']:
        raise e
    self.report_module_time('k -> R')


  def minimal(self,first_band=None, R=False):
      from .defs.do_minimal import do_minimal
      raise Exception('ONLY FOR ARCHIVAL PUTPOSES - DO NOT USE')
      do_minimal(self.data_controller,first_band)
      if R:
        from .defs.do_build_pao_hamiltonian import do_Hks_to_HRs
        do_Hks_to_HRs(self.data_controller)
        self.data_controller.broadcast_single_array('HRs')


  def add_external_fields ( self, Efield=[0.], Bfield=[0.], HubbardU=[0.] ):
    '''
    Add External Fields and Corrections to the Hamiltonian 'HRs'

    Arguments:
        Efield (ndarray): 1D array of Efield values
        Bfield (ndarray): 1D array of Bfield values
        HubbardU (ndarray): 1D array of Hubbard Correctional values

    Returns:
        None
    
    '''
    arry,attr = self.data_controller.data_dicts()

    if any(v != 0. for v in Efield): arry['Efield'] = np.array(Efield)
    if any(v != 0. for v in Bfield): arry['Bfield'] = np.array(Bfield)
    if any(v != 0. for v in HubbardU): arry['HubbardU'] = np.array(HubbardU)

    Efield,Bfield,HubbardU = arry['Efield'],arry['Bfield'],arry['HubbardU']

    try:
      # Add external fields or non scf ACBN0 correction
      if Efield.any() != 0. or Bfield.any() != 0. or HubbardU.any() != 0.:
        from .defs.add_ext_field import add_ext_field
        add_ext_field(self.data_controller)
        if self.rank == 0 and attr['verbose']:
          print('External Fields Added')
    except Exception as e:
      self.report_exception('add_external_fields')
      if attr['abort_on_exception']:
        raise e

    self.comm.Barrier()



  def z2_pack ( self, fname='z2pack_hamiltonian.dat' ):
    '''
    Write 'HRs' to file for use with Z2 Pack

    Arguments:
        fname (str): File name for the Hamiltonian in the Z2 Pack format

    Returns:
        None

    '''
    try:
      self.data_controller.write_z2pack(fname)
    except Exception as e:
      self.report_exception('z2_pack')
      if self.data_controller.data_attributes['abort_on_exception']:
        raise e



  def bands ( self, ibrav=None, band_path=None, high_sym_points=None, spin_orbit=False, fname='bands', nk=500 ):
    '''
    Compute the electronic band structure

    Arguments:
        ibrav (int): Crystal structure (following the specifications of QE)
        band_path (str): A string representing the band path to follow
        high_sym_points (dictionary): A dictionary with symbols of high symmetry points as keys and length 3 numpy arrays containg the location of the symmetry points as values.
        spin_orbit (bool): If True the calculation includes relativistic spin orbit coupling
        fname (str): File name for the band output
        nk (int): Number of k-points to include in the path (High Symmetry points are currently included twice, increasing nk)

    Returns:
        None

    '''
    from .defs.do_bands import do_bands
    from .defs.communication import gather_full

    arrays,attr = self.data_controller.data_dicts()

    if ibrav is not None:
      attr['ibrav'] = ibrav

    if 'ibrav' not in attr and 'kq' not in arrays:
      if band_path is None or high_sym_points is None:
        if self.rank == 0:
          print('Must specify the high-symmetry path, \'kq\', or \'ibrav\'')

    if 'nk' not in attr: attr['nk'] = nk
    if band_path is not None: attr['band_path'] = band_path
    if 'do_spin_orbit' not in attr: attr['do_spin_orbit'] = spin_orbit
    if high_sym_points is not None: arrays['high_sym_points'] = high_sym_points

    # Prepare HRs for band computation with spin-orbit coupling
    try:

      # Calculate the bands
      do_bands(self.data_controller)

      if self.rank == 0 and 'nkpnts' in attr and arrays['kq'].shape[1] == attr['nkpnts']:
        print('WARNING: The bands kpath and nscf calculations have the same size.')
        print('Spin Texture calculation should be performed after \'pao_eigh\' to ensure integration across the entire BZ.\n')

      E_kp = gather_full(arrays['E_k'], attr['npool'])
      self.data_controller.write_bands(fname, E_kp)
      E_kp = None
    except Exception as e:
      self.report_exception('bands')
      if attr['abort_on_exception']:
        raise e

    self.report_module_time('Bands')


  def adhoc_spin_orbit ( self, naw=[1], phi=.0, theta=.0,  lambda_p=[.0], lambda_d=[.0], soc_strengh={}, soc_species=True ):
    '''
    Include spin-orbit coupling  

    Arguments:
        theta (float)             :  Spin orbit angle
        phi (float)               :  Spin orbit azimuthal angle
        soc_strengh(dict)         :  p and d orbitals SOC strengh for each species 
        If soc_species = False
        lambda_p (list of floats) :  p orbitals SOC strengh for each atom 
        lambda_d (list of float)  :  d orbitals SOC strengh for each atom

    Returns:
        None

    '''
    from .defs.do_spin_orbit import do_spin_orbit_H

    arry,attr = self.data_controller.data_dicts()
    attr['do_spin_orbit'] = attr['adhoc_SO'] = True

    if 'phi' not in attr: attr['phi'] = phi
    if 'theta' not in attr: attr['theta'] = theta

    if (soc_species==True):
      lambda_p=[]
      lambda_d=[]
      for i in range (len(arry['atoms'])):
        lambda_p.append(soc_strengh[arry['atoms'][i]][0])
        lambda_d.append(soc_strengh[arry['atoms'][i]][1])
      arry['lambda_p']=lambda_p
      arry['lambda_d']=lambda_d
    else:
      if 'lambda_p' not in arry: arry['lambda_p'] = lambda_p[:]
      if 'lambda_d' not in arry: arry['lambda_d'] = lambda_d[:]

    self.data_controller.build_arrays_adhoc_soc()
    
    # Check if the pseudo potential or internal basis configuraton is implemented
    if (len(arry['orb_pseudo'])==attr['natoms']):
      # add SOC  
      do_spin_orbit_H(self.data_controller)
      # Rezising arrays
      attr['bnd'] *= 2
      attr['dftSO'] = True
      attr['nspin'] = 1
      attr['nawf'] = arry['HRs'].shape[0]
      # for write_z2pack
      del arry["Hks"]
      arry["Hks"]=np.fft.fftn(arry["HRs"],axes=(2,3,4))
    else:
      self.report_module_time('adhoc_spin_orbit')
      raise(NotImplementedError("Pseudo potential or internal basis configuration not implemented"))
    
    self.report_module_time('adhoc_spin_orbit')

  def wave_function_projection ( self, dimension=3 ):
    '''
    Marcio, can you write something here please?
    Please check the argument descriptions also, I half guessed.

    Arguments:
        dimension (int): Dimensionality of the system

    Returns:
        None
mo    '''
    from .defs.do_wave_function_site_projection import wave_function_site_projection

    try:
      wave_function_site_projection(self.data_controller)
    except Exception as e:
      self.report_exception('wave_function_projection')
      if self.data_controller.data_attributes['abort_on_exception']:
        raise e

    self.report_module_time('wave_function_projection')


  def doubling_Hamiltonian ( self, nx , ny, nz ):
    '''
    Marcio, can you write something here please?
    Please check the argument descriptions also

    Arguments:
        nx (bool): Number of doubles in first dimension
        ny (bool): Number of doubles in second dimension
        nz (bool): Number of doubles in third dimension

    Returns:
        None
    '''
    from .defs.do_doubling import doubling_HRs
    
    arrays,attr = self.data_controller.data_dicts()
    attr['nx'],attr['ny'],attr['nz'] = nx,ny,nz
    
    try:
      doubling_HRs(self.data_controller)
    except Exception as e:
      self.report_exception('doubling_Hamiltonian')
      if attr['abort_on_exception']:
        raise e

    self.report_module_time('doubling_Hamiltonian')
  


  def cutting_Hamiltonian ( self, x=False , y=False, z=False ):
    '''
    Marcio, can you write something here please?

    Arguments:
        x (bool): If True, cut along the first dimension
        y (bool): If True, cut along the second dimension
        z (bool): If True, cut along the third dimension

    Returns:
        None
    '''
    arry,attr = self.data_controller.data_dicts()

    try:
      if x:
        for i in range(attr['nk1']-1,0,-1):
          arry['HRs'] = np.delete(arry['HRs'],i,2)
      if y:
        for i in range(attr['nk2']-1,0,-1):
          arry['HRs'] = np.delete(arry['HRs'],i,3)
      if z:
        for i in range(attr['nk3']-1,0,-1):
          arry['HRs'] = np.delete(arry['HRs'],i,4)

      _,_,attr['nk1'],attr['nk2'],attr['nk3'],_ = arry['HRs'].shape
      attr['nkpnts'] = attr['nk1']*attr['nk2']*attr['nk3']
    except Exception as e:
      self.report_exception('cutting_Hamiltonian')
      if attr['abort_on_exception']:
        raise e



  def spin_operator ( self, spin_orbit=False, sh_l=None, sh_j=None):
    '''
    Calculate the Spin Operator for calculations involving spin
      Requires: None
      Yeilds: 'Sj'

    Arguments:
        spin_orbit (bool): If True the calculation includes relativistic spin orbit coupling
        fnscf (string): Filename for the QE nscf inputfile, from which to read shell data
        sh (list of ints): The Shell levels
        nl (list of ints): The Shell level occupations

    Returns:
        None
    '''
    arrays,attr = self.data_controller.data_dicts()

    if 'do_spin_orbit' not in attr: attr['do_spin_orbit'] = spin_orbit
    adhoc_SO = 'adhoc_SO' in attr and attr['adhoc_SO']

    if ('sh_l' not in arrays and 'sh_j' not in arrays) and not adhoc_SO:
      if sh_l is None and sh_j is None:
        sh = arrays['shells']
        shells,jchia = [],[]
        for i,a in enumerate(arrays['atoms']):
          ash = []
          for v in sh[a]:
            ash += [v, v] if v==0 else [v]
          shells += ash[::2]
          for l in ash[::2]:
            jchia += [.5, .5] if l==0 else [l-.5, l+.5]
        arrays['sh_j'] = np.array(jchia)[::2]
        arrays['sh_l'] = np.array(shells)
      else:
        arrays['sh_l'],arrays['sh_j'] = sh_l,sh_j
    try:
      nawf = attr['nawf']

      # Compute spin operators
      # Pauli matrices (x,y,z)
      Sj = np.zeros((3,nawf,nawf), dtype=complex)
      sP = 0.5*np.array([[[0.0,1.0],[1.0,0.0]],[[0.0,-1.0j],[1.0j,0.0]],[[1.0,0.0],[0.0,-1.0]]])
      if spin_orbit:
        # Spin operator matrix  in the basis of |l,m,s,s_z> (TB SO)
        for spol in range(3):
          for i in range(nawf//2):
            Sj[spol,i,i] = sP[spol][0,0]
            Sj[spol,i,i+1] = sP[spol][0,1]
          for i in range(nawf//2, nawf):
            Sj[spol,i,i-1] = sP[spol][1,0]
            Sj[spol,i,i] = sP[spol][1,1]
      else:
        from .defs.clebsch_gordan import clebsch_gordan
        # Spin operator matrix  in the basis of |j,m_j,l,s> (full SO)
        for spol in range(3):
          Sj[spol,:,:] = clebsch_gordan(nawf, arrays['sh_l'], arrays['sh_j'], spol)

      arrays['Sj'] = Sj
    except Exception as e:
      self.report_exception('spin_operator')
      if attr['abort_on_exception']:
        raise e



  def topology ( self, eff_mass=False, Berry=False, spin_Hall=False, spin_orbit=False, spol=None, ipol=None, jpol=None ):
    '''
    Calculate the Band Topology along the k-path 'kq'

    Arguments:
        eff_mass (bool): If True calculate the Effective Mass Tensor
        Berry (bool): If True calculate the Berry Curvature
        spin_Hall (bool): If True calculate Spin Hall Conductivity
        spin_orbit (bool): If True the calculation includes spin_orbit effects for topology.
        spol (int): Spin polarization
        ipol (int): In plane dimension 1
        jpol (int): In plane dimension 2

    Returns:
        None
    '''
    from .defs.do_topology import do_topology
    # Compute Z2 invariant, velocity, momentum and Berry curvature and spin Berry
    # curvature operators along the path in the IBZ from do_topology_calc

    arrays,attr = self.data_controller.data_dicts()

    if 'Berry' not in attr: attr['Berry'] = Berry
    if 'eff_mass' not in attr: attr['eff_mass'] = eff_mass
    if 'spin_Hall' not in attr: attr['spin_Hall'] = spin_Hall
    if 'do_spin_orbit' not in attr: attr['do_spin_orbit'] = spin_orbit

    attr['spol'] = spol
    attr['ipol'] = ipol
    attr['jpol'] = jpol

    if attr['spol'] is None or attr['ipol'] is None or attr['jpol'] is None:
      if self.rank == 0:
        print('Must specify \'spol\', \'ipol\', and \'jpol\'')
      quit()

    if spin_Hall and 'Sj' not in arrays:
      self.spin_operator(spin_orbit=attr['do_spin_orbit'])

    try:
      do_topology(self.data_controller)
    except Exception as e:
      self.report_exception('topology')
      if attr['abort_on_exception']:
        raise e

    self.report_module_time('Band Topology')

    del arrays['R']
    del arrays['idx']
    del arrays['Rfft']
    del arrays['R_wght']



  def interpolated_hamiltonian ( self, nfft1=0, nfft2=0, nfft3=0, reshift_Ef=False ):
    '''
    Calculate the interpolated Hamiltonian with the method of zero padding
    Populates DataController with 'Hksp'

    Arguments:
        nfft1 (int): Desired size of the interpolated Hamiltonian's first dimension
        nfft2 (int): Desired size of the interpolated Hamiltonian's second dimension
        nfft3 (int): Desired size of the interpolated Hamiltonian's third dimension

    Returns:
        None
    '''
    from .defs.get_K_grid_fft import get_K_grid_fft
    from .defs.do_double_grid import do_double_grid
    from .defs.communication import gather_scatter
    from .defs.do_Efermi import E_Fermi

    arrays,attr = self.data_controller.data_dicts()

    try:

      if 'HRs' not in arrays:
        raise KeyError('HRs')

      nawf = attr['nawf']
      nko1,nko2,nko3 = attr['nk1'],attr['nk2'],attr['nk3']

      # Automatically doubles grid in any direction with unspecified nfft value
      if nfft1 == 0: nfft1 = 2*nko1
      if nfft2 == 0: nfft2 = 2*nko2
      if nfft3 == 0: nfft3 = 2*nko3

      attr['nfft1'],attr['nfft2'],attr['nfft3'] = nfft1,nfft2,nfft3

      # Adjust 'npool' if arrays exceed MPI maximum
      int_max = 2147483647
      temp_pool = int(np.ceil((float(nawf**2*nfft1*nfft2*nfft3*3*attr['nspin'])/float(int_max))))
      if temp_pool > attr['npool']:
        if self.rank == 0:
          print("Warning: %s too low. Setting npool to %s"%(attr['npool'],temp_pool))
        attr['npool'] = temp_pool

      # Fourier interpolation on extended grid (zero padding)
      do_double_grid(self.data_controller)

      snawf,_,_,_,nspin = arrays['Hksp'].shape
      arrays['Hksp'] = np.reshape(arrays['Hksp'], (snawf,attr['nkpnts'],nspin))
      arrays['Hksp'] = gather_scatter(arrays['Hksp'], 1, attr['npool'])

      snktot = arrays['Hksp'].shape[1]
      if reshift_Ef:
        Hksp = arrays['Hksp'].reshape((nawf,nawf,snktot,nspin))
        Ef = E_Fermi(Hksp, self.data_controller, parallel=True)
        dinds = np.diag_indices(nawf)
        Hksp[dinds[0], dinds[1]] -= Ef
        arrays['Hksp'] = np.moveaxis(Hksp, 2, 0)
      else:
        arrays['Hksp'] = np.reshape(np.moveaxis(arrays['Hksp'],0,1), (snktot,nawf,nawf,nspin))

      get_K_grid_fft(self.data_controller)

      # Report new memory requirements
      if self.rank == 0:
        gbyte = self.memory_check()
        if attr['verbose']:
          print('Performing Fourier interpolation on a larger grid.')
          print('d : nk -> nfft\n1 : %d -> %d\n2 : %d -> %d\n3 : %d -> %d'%(nko1,nfft1,nko2,nfft2,nko3,nfft3))
        print('New estimated maximum array size: %.2f GBytes'%gbyte)

    except Exception as e:
      self.report_exception('interpolated_hamiltonian')
      if attr['abort_on_exception']:
        raise e

    self.report_module_time('R -> k with Zero Padding')



  def pao_eigh ( self, bval=0 ):
    '''
    Calculate the Eigen values and vectors of k-space Hamiltonian 'Hksp'
    Populates DataController with 'E_k' and 'v_k'

    Arguments:
        bval (int): Top valence band number (nelec/2) to correctly shift Eigenvalues

    Returns:
        None
    '''
    from .defs.do_eigh import do_pao_eigh
    from .defs.communication import gather_scatter,scatter_full,gather_full

    arrays,attr = self.data_controller.data_dicts()

    if 'bval' not in attr: attr['bval'] = bval

    # HRs and Hks are replaced with Hksp
    if 'HRs' in arrays:
      del arrays['HRs']

    try:
      if 'Hksp' not in arrays:
        if self.rank == 0:
          nktot = attr['nkpnts']
          nawf,_,nk1,nk2,nk3,nspin = arrays['Hks'].shape
          arrays['Hks'] = np.moveaxis(np.reshape(arrays['Hks'],(nawf,nawf,nktot,nspin),order='C'), 2, 0)
        else:
          arrays['Hks'] = None
        arrays['Hksp'] = scatter_full(arrays['Hks'], attr['npool'])
        del arrays['Hks']

      do_pao_eigh(self.data_controller)

      ### PARALLELIZATION
      ## DEV: Sample RunTime Here
      ## DEV: Parallelize search for amax & subtract for all processes.
      if 'HubbardU' in arrays and arrays['HubbardU'].any() != 0.0:
        arrays['E_k'] = gather_full(arrays['E_k'], attr['npool'])
        if self.rank == 0:
          if attr['verbose']:
            print('Shifting Eigenvalues to top of valence band.')
          arrays['E_k'] -= np.amax(arrays['E_k'][:,attr['bval'],:])
        self.comm.Barrier()
        arrays['E_k'] = scatter_full(arrays['E_k'], attr['npool'])
    except Exception as e:
      self.report_exception('pao_eigh')
      if attr['abort_on_exception']:
        raise e

    self.report_module_time('Eigenvalues')



  def gradient_and_momenta ( self,band_curvature=False ):
    '''
    Calculate the Gradient of the k-space Hamiltonian, 'Hksp'
    Requires 'Hksp'
    Populates DataController with 'dHksp'

    Arguments:
      None

    Returns:
      None
    '''
    from .defs.do_gradient import do_gradient
    from .defs.do_momentum import do_momentum
    from .defs.communication import gather_scatter
    import numpy as np 
    from .defs.do_atwfc_proj import build_pswfc_basis_all, build_aewfc_basis

    arrays,attr = self.data_controller.data_dicts()

    try:
      snktot,nawf,_,nspin = arrays['Hksp'].shape

      for ik in range(snktot):
        for ispin in range(nspin):
          #make sure Hksp is hermitian (it should be)
          arrays['Hksp'][ik,:,:,ispin] = (np.conj(arrays['Hksp'][ik,:,:,ispin].T) + arrays['Hksp'][ik,:,:,ispin])/2.

      arrays['Hksp'] = np.reshape(arrays['Hksp'], (snktot, nawf**2, nspin))
      arrays['Hksp'] = np.moveaxis(gather_scatter(arrays['Hksp'],1,attr['npool']), 0, 1)
      snawf,_,nspin = arrays['Hksp'].shape
      arrays['Hksp'] = np.reshape(arrays['Hksp'], (snawf,attr['nk1'],attr['nk2'],attr['nk3'],nspin))

      do_gradient(self.data_controller)

      if not band_curvature:
        # No more need for k-space Hamiltonian
        del arrays['Hksp']

      ### PARALLELIZATION
      #gather dHksp on nawf*nawf and scatter on k points
      arrays['dHksp'] = np.reshape(arrays['dHksp'], (snawf,attr['nkpnts'],3,nspin))
      arrays['dHksp'] = np.moveaxis(gather_scatter(arrays['dHksp'],1,attr['npool']), 0, 2)
      arrays['dHksp'] = np.reshape(arrays['dHksp'], (snktot,3,nawf,nawf,nspin), order="C")

      for nk in range(snktot):
        for i in range(3):
          for s in range(nspin):
            arrays['dHksp'][nk,i,:,:,s] = (arrays['dHksp'][nk,i,:,:,s] + np.conj(arrays['dHksp'][nk,i,:,:,s].T))/2.
      if band_curvature:
        from .defs.do_band_curvature import do_band_curvature
        do_band_curvature(self.data_controller)
        # No more need for k-space Hamiltonian
        del arrays['Hksp']
      
    except Exception as e:
      self.report_exception('gradient_and_momenta')
      if attr['abort_on_exception']:
        raise e

    self.report_module_time('Gradient')

    ### DEV: Proposed to remove this and calculate pksp or velkp when required
    # Compute the momentum operator p_n,m(k) (and kinetic energy operator)
    do_momentum(self.data_controller)
    self.report_module_time('Momenta')



  def adaptive_smearing ( self, smearing='gauss' ):
    '''
    Calculate the Adaptive Smearing parameters
    Populates DataController with 'deltakp' and 'deltakp2'

    Arguments:
        smearing (str): Smearing type (m-p and gauss)

    Returns:
        None
    '''
    from .defs.do_adaptive_smearing import do_adaptive_smearing

    attr = self.data_controller.data_attributes

    if smearing != 'gauss' and smearing != 'm-p':
      raise ValueError('Smearing type %s not supported.\nSmearing types are \'gauss\' and \'m-p\''%str(smearing))

    try:
      do_adaptive_smearing(self.data_controller, smearing)
    except Exception as e:
      self.report_exception('adaptive_smearing')
      if attr['abort_on_exception']:
        raise e
    self.report_module_time('Adaptive Smearing')



  def dos ( self, do_dos=True, do_pdos=True, delta=0.01, emin=-10., emax=2., ne=1000 ):
    '''
    Calculate the Density of States and Projected Density of States
      If Adaptive Smearing has been performed, the Adaptive DoS will be calculated

    Arguments:
        do_dos (bool): Perform Density of States calculation
        do_pdos (bool): Perform Projected Density of States calculation
        delta (float): The gaussian width
        emin (float): The minimum energy in the range to be computed
        emax (float): The maximum energy in the range to be computed
        ne (int): The number of points to place in the range [emin,emax]

    Returns:
        None
    '''
    arrays,attr = self.data_controller.data_dicts()

    if 'smearing' not in attr: attr['smearing'] = None

    try:
      if attr['smearing'] is None:
        if do_dos:
          from .defs.do_dos import do_dos
          do_dos(self.data_controller, emin, emax, ne, delta)
        if do_pdos:
          from .defs.do_pdos import do_pdos
          do_pdos(self.data_controller, emin, emax, ne, delta)
      else:
        if 'deltakp' not in arrays:
          if do_dos:
            from .defs.do_dos import do_dos
            do_dos(self.data_controller, emin, emax, ne, delta)
          if do_pdos:
            from .defs.do_pdos import do_pdos
            do_pdos(self.data_controller, emin, emax, ne, delta)
        else:
          if do_dos:
            from .defs.do_dos import do_dos_adaptive
            do_dos_adaptive(self.data_controller, emin, emax, ne)
          if do_pdos:
            from .defs.do_pdos import do_pdos_adaptive
            do_pdos_adaptive(self.data_controller, emin, emax, ne)
    except Exception as e:
      self.report_exception('dos')
      if attr['abort_on_exception']:
        raise e

    mname = 'DoS%s'%('' if attr['smearing'] is None or 'deltakp' not in arrays else ' (Adaptive Smearing)')
    self.report_module_time(mname)



  def density ( self, nr1=48, nr2=48, nr3=48):
    '''
    Calculate the Electron Density in real space
    Arguments:
        nr1,nr2,nr3: real space grid

    Returns:
        None
    '''
    from .defs.do_real_space import do_density

    do_density(self.data_controller, nr1,nr2,nr3)
    
    self.report_module_time('Density')


  
  def trim_non_projectable_bands ( self ):

    arrays, attributes = self.data_controller.data_dicts()

    bnd = attributes['nawf'] = attributes['bnd']

    arrays['E_k'] = arrays['E_k'][:,:bnd]
    arrays['pksp'] = arrays['pksp'][:,:,:bnd,:bnd]
    if 'deltakp' in arrays:
      arrays['deltakp'] = arrays['deltakp'][:,:bnd]
      arrays['deltakp2'] = arrays['deltakp2'][:,:bnd]



  def fermi_surface ( self, fermi_up=1., fermi_dw=-1. ):
    '''
    Calculate the Fermi Surface

    Arguments:
        fermi_up (float): The upper limit of the occupied energy range
        fermi_dw (float): The lower limit of the occupied energy range

    Returns:
        None
    '''
    from .defs.do_fermisurf import do_fermisurf

    attr = self.data_controller.data_attributes

    if 'fermi_up' not in attr: attr['fermi_up'] = fermi_up
    if 'fermi_dw' not in attr: attr['fermi_dw'] = fermi_dw

    try:
      do_fermisurf(self.data_controller)
    except Exception as e:
      self.report_exception('fermi_surface')
      if attr['abort_on_exception']:
        raise e

    self.report_module_time('Fermi Surface')



  def spin_texture ( self, fermi_up=1., fermi_dw=-1. ):
    '''
    Calculate the Spin Texture

    Arguments:
        fermi_up (float): The upper limit of the occupied energy range
        fermi_dw (float): The lower limit of the occupied energy range

    Returns:
        None
    '''
    from .defs.do_spin_texture import do_spin_texture

    arry,attr = self.data_controller.data_dicts()

    if 'fermi_up' not in attr: attr['fermi_up'] = fermi_up
    if 'fermi_dw' not in attr: attr['fermi_dw'] = fermi_dw

    try:
      if attr['nspin'] == 1:
        if 'Sj' not in arry:
          self.spin_operator()
        do_spin_texture(self.data_controller)
        self.report_module_time('Spin Texture')
      else:
        if self.rank == 0:
          print('Cannot compute spin texture with nspin=2')
    except Exception as e:
      self.report_exception('spin_texture')
      if attr['abort_on_exception']:
        raise e


    self.comm.Barrier()



  def spin_Hall ( self, twoD=False, do_ac=False, emin=-1., emax=1., fermi_up=1., fermi_dw=-1., s_tensor=None ):
    '''
    Calculate the Spin Hall Conductivity
      Currently this module does not possess the "spin_orbit" capability of do_topology, because I(Frank) do not know what this modification entails.
      Thus, do_spin_orbit defaults to False here. If anybody needs the spin_orbit capability here, please contact me and we'll sort it out.

    Arguments:
        twoD (bool): True to output in 2D units of Ohm^-1, neglecting the sample height in the z direction
        do_ac (bool): True to calculate the Spic Circular Dichroism
        emin (float): The minimum energy in the range
        emax (float): The maximum energy in the range
        fermi_up (float): The upper limit of the occupied energy range
        fermi_dw (float): The lower limit of the occupied energy range
        s_tensor (list): List of tensor elements to calculate (e.g. To calculate xxx and zxy use [[0,0,0],[0,1,2]])

    Returns:
        None
    '''
    from .defs.do_Hall import do_spin_Hall

    arrays,attr = self.data_controller.data_dicts()

    attr['eminH'],attr['emaxH'] = emin,emax

    if s_tensor is not None: arrays['s_tensor'] = np.array(s_tensor)
    if 'fermi_up' not in attr: attr['fermi_up'] = fermi_up
    if 'fermi_dw' not in attr: attr['fermi_dw'] = fermi_dw

    if 'Sj' not in arrays:
      self.spin_operator()

    try:
      do_spin_Hall(self.data_controller, twoD, do_ac)
    except Exception as e:
      self.report_exception('spin_Hall')
      if attr['abort_on_exception']:
        raise e

    self.report_module_time('Spin Hall Conductivity')


  def rashba_edelstein ( self, emin=-2, emax=2, ne=500, temps=0., reg=1e-30, twoD=False, lt=1., st=1., write_to_file=True ):
    '''
    Calculate the Rashba-Edelstein tensor
    Arguments:
        emin (float): The minimum energy in the range
        emax (float): The maximum energy in the range
        ne (float): The number of energy increments in [emin,emax]
        temps (float): Smearing temperature in eV
        reg (float): The regularization number that is applicable only for materials with band gap (in the order of 1e-30)
        twoD (bool): set True for two dimnesional materials
        lt (float): The lattice height of the twoD structure (in cm)
        st (float): The structure 'effectivce' thickness of the twoD structure (in cm)
        write_to_file (bool): Set True to write tensors to file

    Returns:
        None
    '''
    from .defs.do_rashba_edelstein import do_rashba_edelstein

    arrays,attr = self.data_controller.data_dicts()

    ene = np.linspace(emin, emax, ne)
    try:
      do_rashba_edelstein(self.data_controller, ene, temps, reg, twoD, lt, st, write_to_file)

    except Exception as e:
      self.report_exception('rashba_edelstein')
      if attr['abort_on_exception']:
        raise e

    self.report_module_time('Rashba_Edelstein')


  def anomalous_Hall ( self, do_ac=False, emin=-1., emax=1., fermi_up=1., fermi_dw=-1., a_tensor=None ):
    '''
    Calculate the Anomalous Hall Conductivity

    Arguments:
        do_ac (bool): True to calculate the Magnetic Circular Dichroism
        emin (float): The minimum energy in the range
        emax (float): The maximum energy in the range
        fermi_up (float): The upper limit of the occupied energy range
        fermi_dw (float): The lower limit of the occupied energy range
        a_tensor (list): List of tensor elements to calculate (e.g. To calculate xx and yz use [[0,0],[1,2]])

    Returns:
        None
    '''
    from .defs.do_Hall import do_anomalous_Hall

    arrays,attr = self.data_controller.data_dicts()

    attr['eminH'] = emin
    attr['emaxH'] = emax

    if a_tensor is not None: arrays['a_tensor'] = np.array(a_tensor)
    if 'fermi_up' not in attr: attr['fermi_up'] = fermi_up
    if 'fermi_dw' not in attr: attr['fermi_dw'] = fermi_dw

    try:
      do_anomalous_Hall(self.data_controller, do_ac)
    except Exception as e:
      self.report_exception('anomalous_Hall')
      if attr['abort_on_exception']:
        raise e

    self.report_module_time('Anomalous Hall Conductivity')


  def effective_mass ( self, emin=-1., emax=1., ne=1000 ):

    '''
    Calculate effective mass components for kx,ky and kz

    Arguments:
        tmin (float): The minimum temperature in the range
        tmax (float): The maximum temperature in the range
        nt (float): The number of temperature increments
        emin (float): The minimum energy in the range
        emax (float): The maximum energy in the range
        ne (float): The number of energy increments

    Returns:
        None
    '''
    from .defs.do_effective_mass import do_effective_mass

    arrays,attr = self.data_controller.data_dicts()

    ene = np.linspace(emin, emax, ne)

    try:
      do_effective_mass(self.data_controller, ene)

    except Exception as e:
      self.report_exception('effective_mass')
      if attr['abort_on_exception']:
        raise e

    self.report_module_time('Effective mass')



  def doping ( self, tmin=300, tmax=300, nt=1, delta=0.01, emin=-1., emax=1., ne=1000, doping_conc=0., core_electrons=0., fname='doping_' ):
    '''
    Calculate the chemical potential that corresponds to specified doping for different temperatures

    Arguments:
        tmin (float): The minimum temperature in the range
        tmax (float): The maximum temperature in the range
        nt (float): The number of temperature increments
        emin (float): The minimum energy in the range
        emax (float): The maximum energy in the range
        ne (float): The number of energy increments
        doping_conc(float): The amount of doping in 1/cm^3. Positive value for p type and negative value for n type
        core_electrons(float): The number of core electrons in a system. Adding this to integrated dos allows for integration over a narrower energy window
        fname (str): Prefix for the filename containg Doping vs Temperature

    Returns:
        None
    '''
    from .defs.do_doping import do_doping
    from .defs.do_dos import do_dos,do_dos_adaptive
    
    arrays,attr = self.data_controller.data_dicts()

    if 'delta' not in attr: attr['delta'] = delta
    if 'doping_conc' not in attr: attr['doping_conc'] = doping_conc
    if 'core_electrons' not in attr: attr['core_electrons'] = core_electrons

    ene = np.linspace(emin, emax, ne)
    temps = np.linspace(tmin, tmax, nt)
    if attr['smearing'] == None:
      do_dos(self.data_controller, emin, emax, ne, delta)
    else:
      do_dos_adaptive(self.data_controller, emin, emax, ne, delta)
    do_doping(self.data_controller, temps, ene, fname)

    self.report_module_time('Doping')



  def transport ( self, tmin=300., tmax=300., nt=1, emin=-2., emax=2., ne=500, scattering_channels=[], scattering_weights=[], tau_dict={}, do_hall=False, write_to_file=True, save_tensors=False ):
    '''
    Calculate the Transport Properties

    Arguments:
        tmin (float): The minimum temperature in the range
        tmax (float): The maximum temperature in the range
        nt (float): The number of temperatures to evaluate in [tmin,tmax]
        emin (float): The minimum energy in the range
        emax (float): The maximum energy in the range
        ne (float): The number of energy increments in [emin,emax]
        fit (bool): fit paoflow output to experiments
        scattering_channels (list): List of strings specifying a built in model and/or TauModel objects to evaluate
        scattering_weights (list): List of coefficients weighting components of the harmonic sum when computing tau
        tau_dict (dict): Dictionary housing parameters required for calculating Tau with built-in models
        do_hall (bool): Set True to calculate hall coefficient
        write_to_file (bool): Set True to write tensors to file
        save_tensors (bool): Set True to save the tensors into the data controller

    Returns:
        None
    '''
    from .defs.do_transport import do_transport

    arrays,attr = self.data_controller.data_dicts()
    if 'tau_dict' not in attr: attr['tau_dict'] = tau_dict

    ene = np.linspace(emin, emax, ne)
    temps = np.linspace(tmin, tmax, nt)
    sc,sw = scattering_channels,scattering_weights
    try:
      # Compute Velocities for Spin 0 Only
      bnd = attr['bnd']
      velkp = np.zeros((arrays['pksp'].shape[0],3,bnd,attr['nspin']))
      for n in range(bnd):
        velkp[:,:,n,:] = np.real(arrays['pksp'][:,:,n,n,:])

      do_transport(self.data_controller, temps, ene, velkp, sc, sw, do_hall, write_to_file, save_tensors)

    except Exception as e:
      self.report_exception('transport')
      if attr['abort_on_exception']:
        raise e

    self.report_module_time('Transport')


  def dielectric_tensor ( self, delta=0.1, intrasmear=0.05, emin=0., emax=10., ne=501, d_tensor=None, degauss=0.1, from_wfc=None):
    '''
    Calculate the Dielectric Tensor

    Arguments:
        delta (float): Inter-smearing parameter in eV
        intrasmear (float): Intra-smearing parameter for metal in eV
        emin (float): The minimum value of energy
        emax (float): The maximum value of energy
        ne (float): Number of energy values between emin and emax
        d_tensor (list): List of tensor elements to calculate (e.g. To calculate xx and yz use [[0,0],[1,2]])
        Can also use options 'all', 'diag', 'offdiag'.

    Returns:
        None
    '''

    from .defs.do_epsilon import do_dielectric_tensor

    arrays,attr = self.data_controller.data_dicts()

    if 'degauss' not in attr: attr['degauss'] = degauss
    if 'delta' not in attr: attr['delta'] = delta
    attr['intrasmear'] = intrasmear
    if d_tensor == 'all':
      pass
    elif d_tensor == 'diag':
      arrays['d_tensor'] = np.array([[0,0],[1,1],[2,2]])
    elif d_tensor == 'offdiag':
      arrays['d_tensor'] = np.array([[0,1],[1,0],[0,2],[2,0],[1,2],[2,1]])
    else:
      arrays['d_tensor'] = np.array(d_tensor)

    #-----------------------------------------------
    # Compute dielectric tensor (Re and Im epsilon)
    #-----------------------------------------------
    try:
      ene = np.linspace(emin, emax, ne)
      do_dielectric_tensor(self.data_controller, ene, from_wfc)
    except Exception as e:
      self.report_exception('dielectric_tensor')
      if attr['abort_on_exception']:
        raise e

    self.report_module_time('Dielectric Tensor')

  def jdos (self, delta=0.1, emin=0., emax=10., ne=501, jdos_smeartype='gauss'):
    '''
    Calculate the Dielectric Tensor

    Arguments:
        delta (float): broadening parameter in eV
        emin (float): The minimum value of energy
        emax (float): The maximum value of energy
        ne (float): Number of energy values between emin and emax
        jdos_smeartype: 'gauss' or "lorentz"

    Returns:
        None
    '''
    from .defs.do_epsilon import do_jdos
    _,attr = self.data_controller.data_dicts()
    if 'delta' not in attr: attr['delta'] = delta

    try:
      ene = np.linspace(emin, emax, ne)
      do_jdos(self.data_controller, ene, jdos_smeartype)
    except Exception as e:
      self.report_exception('joint density of states')
      if attr['abort_on_exception']:
        raise e

    self.report_module_time('Joint density of states')


  def find_weyl_points ( self, symmetrize=None, test_rad=0.01, search_grid=[8,8,8] ):
    from .defs.do_find_Weyl import find_weyl

    try:
      if symmetrize is not None:
        self.data_controller.data_attributes['symmetrize'] = symmetrize
      find_weyl(self.data_controller, test_rad, search_grid)

    except Exception as e:
      self.report_exception('pao_hamiltonian')
      if self.data_controller.data_attributes['abort_on_exception']:
        raise e


    self.report_module_time('Weyl Search')

  def ipr(self, fname='ipr'):
    '''
    Compute the inverse partiticipation ratio (IPR) from PAO eigenstates

                 \sum_n |v_nk|^4
    IPR_nk = -----------------------
              ( \sum_n |v_nk|^2 )^2

    where n is the band index and k the k-point

    The final shape is (nspin,nkpts,nbands,3),
    where the last axis gives: 0 as the k-point coordinate,
                               1 the band energy, and
                               2 the inverse partition ratio

    The result in saved to a ipr.npy file.
    To open the file one should use:
    `ipr = np.load('ipr.npy')`

    '''

    from .defs.do_ipr import inverse_participation_ratio
    from os.path import join

    arry, attr = self.data_controller.data_dicts()

    try:
      arry['ipr'] = inverse_participation_ratio(self.data_controller)
      np.save(join(attr['opath'],fname+'.npy'),arry['ipr'])

    except Exception as e:
      self.report_exception('ipr')
      if attr['abort_on_exception']:
        raise e

    self.report_module_time('Inverse Participation Ratio (IPR)')

  def berry_phase (self, kspace_method='path', berry_path=None, high_sym_points=None, kpath_funct=None, nk1=100, nk2=100, closed=True, method='berry', sub=None, occupied=True, kradius=None, kcenter=[0.,0.,0.],kxlim=(-0.5,0.5),kylim=(-0.5,0.5),eigvals=False,fname='berry_phase',contin=False):
    '''
    Compute the Berry phase using the discretized formula.
    See R. Resta, Rev. Mod. Phys. 66, 899 (1994) and R. Resta, J. Phys.: Condens. Matter 22, 123201 (2010)

    Arguments:
        kspace_method (str): method used to sample the BZ:
                             *'path': 1D path along the BZ;
                             *'track': 1D path along x direction for several points in the y direction;
                             *'circle': circular path around a k-point, given center and radius;
                             *'square': retangular region with nk1 points along x and nk2 points along y. Region defined given x and y start and end points or full BZ.
        berry_path (str): A string representing the band path to follow. The first and last k-point must not be the same.
        berry_high_sym_points (dictionary): A dictionary with symbols of high symmetry points as keys and length 3 numpy arrays containg the location of the symmetry points as values.
        nk (int): Number of k-points to include in the path
        closed (bool, optional): whether or not to include the connection of the last and first points in the loop
        method (str, {'berry','zak'}): 'berry' returns the usual berry phase. 'zak' includes the Zak phase for 1D systems which takes into account the Bloch factor exp(-iG.r)
                                        accumulated over a Brillouin zone. See J. Zak, Phys. Rev. Lett. 62, 2747 (1989)
        sub (None or list of int, optional): index of selected bands to calculate the Berry phase
        occupied (bool, optional): calculate the Berry phase over all occupied bands (if set to True, sub is set to None)
        kxlim (tuple, float): start and end points in x direction for sampling the BZ, used when kspace_method='square' .
        kylim (tuple, float): start and end points in y direction for sampling the BZ, used when kspace_method='square' .
        eigvals (bool, optional): write overlap matrix eigenvalues
        eigvecs (bool, optional): write overlap matrix eigenvectors
        save_prd (bool, optional): save overlap matrix
        fname (str, optional): filename to save the berry phase results
        contin (bool, optional): make berry phase continuous by fixing 2pi shifts

    Returns:
        Berry/Zak phase

    '''
    from .defs.do_berry_phase import do_berry_phase

    arry,attr = self.data_controller.data_dicts()

    kspace_method = kspace_method.lower()
    attr['berry_kspace_method'] = kspace_method

    attr['berry_path'] = berry_path
    arry['berry_high_sym_points'] = high_sym_points

    attr['berry_nk'] = nk1
    attr['berry_nk1'] = nk1
    attr['berry_nk2'] = nk2

    attr['berry_kpath_funct'] = kpath_funct

    arry['berry_kxlim'] = kxlim
    arry['berry_kylim'] = kylim

    if kspace_method != 'square':
      attr['berry_eigvals'] = eigvals
    else:
      attr['berry_eigvals'] = False

    attr['berry_kradius'] = kradius
    arry['berry_kcenter'] = np.array(kcenter)
    attr['berry_eigvals'] = eigvals

    method = method.lower()
    if method in ['berry','zak']:
      attr['berry_method'] = method
    else:
      print('method should be either berry or zak. Falling back to method = \'berry\'')
      attr['berry_method'] = 'berry'

    arry['berry_sub'] = sub
    if sub != None or (sub == None and not occupied):
      attr['berry_occupied'] = False
    else:
      attr['berry_occupied'] = occupied

    attr['berry_closed'] = closed
    attr['berry_contin'] = contin
    attr['berry_fname'] = fname

    try:

      do_berry_phase(self)

    except Exception as e:
      self.report_exception('berry_phase')
      if attr['abort_on_exception']:
        raise e

    self.report_module_time('Berry phase')
